"""The functions used for reading ncaa db files.

The core function of this module, `read_db()`, is the main method for allowing users to
open an ncaa db file. The function takes an opened `BinaryIO` stream and returns a
`NcaaDbFile` class object containing file header information and the db's total tabular
data.

Usage example:

    with open(filename, "wb") as ncaa_db_file:
        file = ncaadb.read_file(ncaa_db_file)

    players_table = file["PLAY"]
"""

import struct
from typing import BinaryIO

import pandas as pd

from ncaadb.const import (
    FILE_HEADER_SIZE,
    TABLE_DEFINITION_SIZE,
    TABLE_FIELD_SIZE,
    TABLE_HEADER_SIZE,
)
from ncaadb.file import Field, FileHeader, NcaaDbFile, Table, TableHeader


class MissingHeaderError(Exception):
    """Exception raised when trying to access a missing header attribute."""


def read_file_header(db_file: BinaryIO) -> FileHeader:
    """Reads the file header from the db file and returns it as a FileHeader object.

    Args:
        db_file (BinaryIO): Open file stream to NCAA DB file

    Returns:
        FileHeader: FileHeader object containing file header information
    """
    buffer = db_file.read(FILE_HEADER_SIZE)
    return FileHeader(*struct.unpack(">HHIIIII", buffer))


def read_table_definitions(db_file: BinaryIO, table_count: int) -> dict[str, Table]:
    """Reads the table definitions from the db file and stores it in a dict.

    Args:
        db_file (BinaryIO): Open file stream to NCAA DB file
        table_count (int): Number of tables in the db file

    Returns:
        dict[str, Table]: Dict mapping table name to table object
    """
    tables = {}
    for _ in range(table_count):
        buffer = db_file.read(TABLE_DEFINITION_SIZE)
        name, offset = struct.unpack(">4sI", buffer)
        name = name.decode()[::-1]
        tables[name] = Table(name, offset)
    return tables


def read_table_fields(db_file: BinaryIO, table: Table) -> None:
    """Reads the table fields from the db file and stores it in the table object.

    Args:
        db_file (BinaryIO): Open file stream to NCAA DB file
        table (Table): Table object to store the records in

    Raises:
        MissingHeaderError: Raised when trying to read from table without header
    """
    if table.header is None:
        raise MissingHeaderError

    for _ in range(table.header.num_fields):
        buffer = db_file.read(TABLE_FIELD_SIZE)
        field = Field(*struct.unpack(">II4sI", buffer))
        table.fields.append(field)


def read_table_records(db_file: BinaryIO, table: Table) -> None:
    """Reads the table records from the db file and stores it in the table object.

    Args:
        db_file (BinaryIO): Open file stream to NCAA DB file
        table (Table): Table object to store the records in

    Raises:
        MissingHeaderError: Raised when trying to read from table without header
    """
    if table.header is None:
        raise MissingHeaderError

    records = []
    for _ in range(table.header.current_records):
        buffer = db_file.read(table.header.len_bytes)
        records.append(buffer)

    columns = [field.name for field in table.fields]
    data = [
        [
            field.read_func(buffer, field.bits, field.offset)
            for field in table.fields
            if callable(field.read_func)
        ]
        for buffer in records
    ]
    table.data = pd.DataFrame(data, columns=columns)


def read_table_data(db_file: BinaryIO, tables: dict[str, Table]) -> None:
    """Reads the table data from the db file and stores it in the table object.

    Args:
        db_file (BinaryIO): Open file stream to NCAA DB file
        tables (dict[str, Table]): Dict mapping table name to table object
    """
    header_start_byte = db_file.tell()
    for table in tables.values():
        bytes_to_skip = (header_start_byte + table.offset) - db_file.tell()
        db_file.read(bytes_to_skip)

        buffer = db_file.read(TABLE_HEADER_SIZE)
        table.header = TableHeader(*struct.unpack(">IIIIIHHIBBHII", buffer))
        read_table_fields(db_file, table)
        read_table_records(db_file, table)


def read_db(db_file: BinaryIO) -> NcaaDbFile:
    """Read an NCAA DB file into python-readable data.

    Args:
        db_file (BinaryIO): Open file stream to NCAA DB file

    Returns:
        NcaaDbFile: NCAA DB File object containing header info and table data
    """
    file_header = read_file_header(db_file)
    table_dict = read_table_definitions(db_file, file_header.table_count)
    file = NcaaDbFile(file_header, table_dict)

    read_table_data(db_file, file.table_dict)
    return file



"""Init file for ncaadb."""

from ncaadb.file import Field, FieldType, FileHeader, NcaaDbFile, Table, TableHeader
from ncaadb.read import read_db

__all__ = [
    "Field",
    "FieldType",
    "FileHeader",
    "NcaaDbFile",
    "Table",
    "TableHeader",
    "read_db",
]



"""The classes comprising the Ncaa DB File representation as `File`."""

from collections.abc import Callable
from dataclasses import dataclass, field
from enum import IntEnum

import pandas as pd

import ncaadb.hex


class FieldType(IntEnum):
    """Enum for the different data types of fields in the Ncaa DB File."""

    STRING = 0
    BINARY = 1
    SINT = 2
    UINT = 3
    FLOAT = 4


@dataclass
class Field:
    """Class to represent the tables' fields in the Ncaa DB File."""

    type: int
    offset: int
    name: bytes | str
    bits: int
    read_func: Callable[[bytes, int, int], int | str] | None = None

    def __post_init__(self) -> None:
        """Post init method to decode the name and convert the type to FieldType."""
        if isinstance(self.name, bytes):
            self.name = self.name.decode()[::-1]

        self.type = FieldType(self.type)
        match self.type:
            case FieldType.STRING:
                self.read_func = ncaadb.hex.read_string
            case FieldType.BINARY:
                self.read_func = ncaadb.hex.read_bytes
            case _:
                self.read_func = ncaadb.hex.read_nums


@dataclass
class TableHeader:
    """Class to store the tables' headers info from the Ncaa DB File."""

    prior_crc: int
    unknown_2: int
    len_bytes: int
    len_bits: int
    zero: int
    max_records: int
    current_records: int
    unknown_3: int
    num_fields: int
    index_count: int
    zero_2: int
    zero_3: int
    header_crc: int


@dataclass
class Table:
    """Class to represent the tables in the Ncaa DB File."""

    name: str
    offset: int
    header: TableHeader | None = None
    fields: list[Field] = field(default_factory=list)
    data: pd.DataFrame | None = None


@dataclass
class FileHeader:
    """Class to store the file header info from the Ncaa DB File."""

    digit: int
    version: int
    unknown_1: int
    db_size: int
    zero: int
    table_count: int
    unknown_2: int


@dataclass
class NcaaDbFile:
    """Class to represent the Ncaa DB File."""

    header: FileHeader
    table_dict: dict[str, Table]

    def __getitem__(self, table_name: str) -> pd.DataFrame:
        """Get the table data from the table name."""
        if table_name not in self.table_dict:
            missing_table_message = f"Table '{table_name}' not found in file"
            raise KeyError(missing_table_message)

        data = self.table_dict[table_name].data
        if data is None:
            missing_data_message = f"Table '{table_name}' has no data"
            raise ValueError(missing_data_message)

        return data

    def __setitem__(self, table_name: str, table_data: pd.DataFrame) -> None:
        """Set the table data for the table name.

        Args:
            table_name (str): The name of the table to set the data for
            table_data (pd.DataFrame): The data to set for the table
        """
        self.table_dict[table_name].data = table_data



"""Functions used to read hex values from hex records.

The tabular data found in Ncaa DB Files is saved for each row in an uninterrupted block
of hex data. Where the data is split for each column and the output data type is found
in the `Field` data. This data along with the block of data is used to read a given row
and column's record.
"""


def read_string(data: bytes, bits: int, offset: int) -> str:
    """Reads tabular record data into a string.

    The strings saved in the hex record data uses blank bits that must be replaced with
    empty strings when decoded in order to maintain proper representation.

    Args:
        data (bytes): A tabular data row's hex record data
        bits (int): The field's data length in bits
        offset (int): The field's data offset to the start in record data

    Returns:
        str: String representation of the row's field-specefic record
    """
    start_byte = offset // 8
    end_byte = (offset + bits) // 8
    return data[start_byte:end_byte].decode("latin-1").replace("\00", "")


def read_bytes(data: bytes, bits: int, offset: int) -> str:
    """Reads tabular record data into a string represenation of hex bytes.

    Args:
        data (bytes): A tabular data row's hex record data
        bits (int): The field's data length in bits
        offset (int): The field's data offset to the start in record data

    Returns:
        str: _description_
    """
    start_byte = offset // 8
    end_byte = (offset + bits) // 8
    return data[start_byte:end_byte].hex()


def read_nums(data: bytes, bits: int, offset: int) -> int:
    """Reads tabular record data into an integer.

    Args:
        data (bytes): A tabular data row's hex record data
        bits (int): The field's data length in bits
        offset (int): The field's data offset to the start in record data

    Returns:
        int: Integer representation of the row's field-specefic record
    """
    byte_offset = offset // 8
    bit_offset = offset % 8
    value = 0
    for _ in range(bits):
        value <<= 1
        value |= (data[byte_offset] >> (7 - bit_offset)) & 1
        bit_offset += 1
        if bit_offset >= 8:
            byte_offset += 1
            bit_offset = 0
    return value
